class CoreException(Exception):
    def __init__(self, failure_msg):
        self.failure_msg = failure_msg
        super().__init__(self.failure_msg)

    pass


class LeaderBoardDoesNotExist(CoreException):
    def __init__(self, leaderboard_id):
        self.failure_msg = f'Leaderboard with id: {leaderboard_id} does not exists.'
        super().__init__(self.failure_msg)


class ProfileIdDoesNotExist(CoreException):
    def __init__(self, profile_id):
        self.failure_msg = f'player_id with id: {profile_id} does not exists.'
        super().__init__(self.failure_msg)


class BadSteamIdExpression(CoreException):
    def __init__(self):
        self.failure_msg = 'Failed to set steam id player. Revise profile_params'
        super().__init__(self.failure_msg)


class BadRelicIdExpression(CoreException):
    def __init__(self):
        self.failure_msg = 'Failed to set relic id player. Revise type of ids'
        super().__init__(self.failure_msg)


class BadAliasesExpression(CoreException):
    def __init__(self):
        self.failure_msg = 'Failed to set alias player. Revise type of alias. Maybe you want' \
                           ' relic id mode? Set to mode="relic"'
        super().__init__(self.failure_msg)


class QueryModeException(CoreException):
    def __init__(self, param):
        self.failure_msg = f'Uknown query mode {param}. Options are ["steam", "relic", "alias"]'
        super().__init__(self.failure_msg)



from enum import Enum


class Codes(Enum):
    """
    Shortcut for leaderboards ID

    USF => American

    UKF => British

    GER => German

    DAK => Afrika Korps
    """
    USF1v1 = 2130255
    UKF1v1 = 2130257
    GER1v1 = 2130261
    DAK1v1 = 2130259

    USF2v2 = 2130300
    UKF2v2 = 2130302
    GER2v2 = 2130306
    DAK2v2 = 2130306

    USF3v3 = 2130329
    UKF3v3 = 2130331
    GER3v3 = 2130335
    DAK3v3 = 2130334

    USF4v4 = 2130353
    UKF4v4 = 2130356
    GER4v4 = 2130360
    DAK4v4 = 2130358


class SortType(Enum):
    """
    Shortcut for Type filter

    ELO => 0
    WINS => 1
    """
    ELO = 0
    WINS = 1



from dataclasses import dataclass, field

from cohpy.endpoint import (
    AllLeaderboards,
    Leaderboard,
    MatchHistory,
    PersonalStats,
)

from .leaderboards import Codes, SortType


@dataclass
class APIClient:
    """
    API interface between user and implementation
    """
    _all_leaderboards: AllLeaderboards = field(default_factory=lambda: AllLeaderboards())
    _specific_leaderboard: Leaderboard = field(default_factory=lambda: Leaderboard())
    _match_history: MatchHistory = field(default_factory=lambda: MatchHistory())
    _personal_stats: PersonalStats = field(default_factory=lambda: PersonalStats())

    def leaderboards(self, *, remove_server_status=True) -> dict:
        """
        Return all types of leaderboards with info

        :param remove_server_status: Set to True if you want the server status response.
        :return: All leaderboards
        """
        response = self._all_leaderboards.get()
        if remove_server_status:
            response.pop('result')
        return response

    def leaderboard(self, *, leaderboard_id, remove_server_status=True, count=200,
                    sort_type=SortType.ELO, start=1):
        """
        Retrieve data about a specific leaderboard given her id.

        :param start: Position of the first player in the requests
        :param sort_type: 1 == Sort by Wins, 0 == Sort by ELO. int or Type instance
        :param count: How many players will be returned [1-200]
        :param leaderboard_id: int or cohpy.leaderboards.Code
        :param remove_server_status: Set to False if you want the server status response.
        :return: leaderboard info dict
        """
        if isinstance(leaderboard_id, Codes):
            leaderboard_id = leaderboard_id.value
        if isinstance(sort_type, SortType):
            sort_type = sort_type.value
        self._specific_leaderboard.query_params['leaderboard_id'] = leaderboard_id
        self._specific_leaderboard.query_params['count'] = count
        self._specific_leaderboard.query_params['type'] = sort_type
        self._specific_leaderboard.query_params['start'] = start
        self._specific_leaderboard.leaderboard_id = leaderboard_id
        response = self._specific_leaderboard.get()

        if remove_server_status:
            response.pop('result')
        return response

    def match_history(self, *, profile_params, remove_server_status=True, mode='relic'):
        """
        :param mode: Query mode against the API. Choose between steam, relic or alias
        :param profile_params: Relic's player (int) id, steam profile (str),
         list of Relic's players ids or list of steam profiles (list)

        single relic's id => profile_params = 1
        single steam id => profile_params = steam/123456789
        single relic's id => profile_params = [1,2,3,4,5...]
        single relic's id => profile_params = [steam/123456789,steam/9786756453423,steam/987654321]

        :param remove_server_status: Set to False if you want the server status response.
        :return:
        """
        self._match_history.profile_params = profile_params
        self._match_history.query_mode = mode
        response = self._match_history.get()
        if remove_server_status:
            response.pop('result')
        return response

    def personal_stats(self, *, profile_params, remove_server_status=True, mode='relic'):
        self._personal_stats.profile_params = profile_params
        self._personal_stats.query_mode = mode
        response = self._personal_stats.get()

        if remove_server_status:
            response.pop('result')
        return response


def get_api_client() -> APIClient:
    """
    :return: APIClient instance
    """
    return APIClient()



import abc
import json
import re
from collections.abc import Iterable
from dataclasses import dataclass

import requests

from .constants import (
    BASE_ACTIONS,
    BASE_COH3_URL,
    TITLE_QUERY_PARAM
)
from .exceptions import (
    BadAliasesExpression,
    BadRelicIdExpression,
    BadSteamIdExpression,
    LeaderBoardDoesNotExist,
    ProfileIdDoesNotExist,
    QueryModeException,
)


class Endpoint(abc.ABC):
    action: str
    title: str = TITLE_QUERY_PARAM
    query_params: dict = {}
    base_actions: str = BASE_ACTIONS
    base_url: str = BASE_COH3_URL

    @property
    def url(self):
        return self._build_url()

    def get(self, **kwargs) -> dict:
        response = requests.get(self.url)
        if not self.validate_response(response):
            raise LeaderBoardDoesNotExist(self.query_params['leaderboard_id'])
        return response.json()

    def _build_url(self) -> str:
        url = f'{self.base_url}{self.base_actions}{self.action}'
        url += self.title
        for param in self.query_params:
            url += f'&{param}={self.query_params.get(param)}'
        return url

    @staticmethod
    def validate_response(response) -> bool:
        """

        :param response: Payload from the API
        :return: False if validation fails else True
        """
        if response.status_code in [400]:
            return False
        response = response.json()
        if response.get('result')['code'] == 5:
            return False
        if response.get('matchHistoryStats') is not None and response.get('profiles') is not None:
            return bool(response.get('matchHistoryStats')) or bool(response.get('profiles'))
        return True


class PlayersEndpoint(Endpoint):
    _profile_params = None
    _mode = None

    @property
    def profile_params(self):
        return self._profile_params

    @profile_params.setter
    def profile_params(self, value):
        self._profile_params = value

    @property
    def query_mode(self):
        return self._mode

    @query_mode.setter
    def query_mode(self, value):
        self._mode = value

    def get(self, **kwargs) -> dict:
        self._set_params()
        response = requests.get(self.url)
        if not self.validate_response(response):
            raise ProfileIdDoesNotExist(self.profile_params)
        return response.json()

    def _set_params(self):
        if self.query_mode not in ['steam', 'relic', 'alias']:
            raise QueryModeException(self.query_mode)
        if self.query_mode == 'steam':
            self._validate_steam_params()
            self.query_params['profile_names'] = json.dumps(self.profile_params)
        elif self.query_mode == 'relic':
            self._validate_relic_params()
            self.query_params['profile_ids'] = json.dumps(self.profile_params)
        elif self.query_mode == 'alias':
            self._validate_aliases_params()
            self.query_params['aliases'] = json.dumps(self.profile_params)

    def _validate_steam_params(self):
        """
        Validate that all steam profiles id are str and startswith steam/

        Regex explanation:

        - ^ is an anchor that specifies the beginning of the string.
        - This means that the string being matched must start with /steam/.
        - [0-9] is a character set that matches any digit (0-9).
        - (+) quantifier specifies that the digit character set must occur one or more times.

         Examples:

        - /steam/123
        - /steam/0
        - /steam/9876543210
        :return:
        """
        self.profile_params = self.profile_params.split() if type(self.profile_params) is str \
            else self.profile_params
        pattern = re.compile(r'^/steam/[0-9]+')
        if not all(type(param) is str and re.fullmatch(pattern, param)
                   for param in self.profile_params):
            raise BadSteamIdExpression()

    def _validate_relic_params(self):
        """
        Validate all relic's params are int
        :return:
        """
        self.profile_params = [self.profile_params] if type(self.profile_params) is not list else \
            self.profile_params
        if not all(type(param) == int for param in self.profile_params):
            raise BadRelicIdExpression()

    def _validate_aliases_params(self):
        """
        Validate all aliases params are str
        :return:
        """
        self.profile_params = self.profile_params.split() if type(self.profile_params) is str \
            else self.profile_params
        if isinstance(self.profile_params, Iterable):
            if not all(type(param) == str for param in self.profile_params):
                raise BadAliasesExpression()
        else:
            if type(self.profile_params) != str:
                raise BadAliasesExpression()


@dataclass
class AllLeaderboards(Endpoint):
    """
    Return all the available leaderboards
    """
    action: str = 'leaderboard/getAvailableLeaderboards/'


@dataclass
class Leaderboard(Endpoint):
    """
    Return concrete data about a leaderboard
    """
    action: str = 'leaderboard/getleaderboard2'


@dataclass
class MatchHistory(PlayersEndpoint):
    """
    Return player match history
    """
    action: str = 'leaderboard/getRecentMatchHistory'


@dataclass
class PersonalStats(PlayersEndpoint):
    """
    Return personal player stats
    """
    action: str = 'leaderboard/getPersonalStat'



from unittest import TestCase

import cohpy


class TestAllLeaderboardsEndpoint(TestCase):

    def setUp(self) -> None:
        self.api_client = cohpy.get_api_client()

    def test_all_leaderboard_with_server_response(self):
        response = self.api_client.leaderboards(remove_server_status=False)
        expected_keys = ('result', 'leaderboards', 'matchTypes', 'races', 'factions',
                         'leaderboardRegions',)

        assert all(key in response for key in expected_keys)
        assert isinstance(response, dict)

    def test_all_leaderboard_without_server_response(self):
        response = self.api_client.leaderboards()
        expected_keys = ('leaderboards', 'matchTypes', 'races', 'factions', 'leaderboardRegions',)

        assert all(key in response for key in expected_keys)
        assert isinstance(response, dict)



import unittest

import cohpy
from cohpy import SortType
from cohpy.exceptions import (
    LeaderBoardDoesNotExist
)


class TestSpecificLeaderboardEndpoint(unittest.TestCase):

    def setUp(self) -> None:
        self.api_client = cohpy.get_api_client()

    def test_specific_leaderboard_with_server_response(self):
        response = self.api_client.leaderboard(leaderboard_id=cohpy.Codes.USF3v3,
                                               remove_server_status=False)
        expected_keys = ('result', 'statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)
        assert isinstance(response, dict)

    def test_all_leaderboard_without_server_response(self):
        response = self.api_client.leaderboard(leaderboard_id=cohpy.Codes.USF3v3)
        expected_keys = ('statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)
        assert isinstance(response, dict)

    def test_all_leaderboard_with_int(self):
        response = self.api_client.leaderboard(leaderboard_id=2130329)
        expected_keys = ('statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)
        assert isinstance(response, dict)

    def test_non_existing_leaderboard_raise_exception(self):
        with self.assertRaises(LeaderBoardDoesNotExist):
            self.api_client.leaderboard(leaderboard_id=-1)

    def test_default_player_len_is_200(self):
        response = self.api_client.leaderboard(leaderboard_id=cohpy.Codes.USF3v3)
        expected_keys = ('statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)
        assert len(response['statGroups']) == 200

    def test_player_count_filter(self):
        response = self.api_client.leaderboard(leaderboard_id=cohpy.Codes.USF3v3, count=50)
        expected_keys = ('statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)
        assert len(response['statGroups']) == 50

    def test_players_sort_by_type(self):
        response = self.api_client.leaderboard(leaderboard_id=cohpy.Codes.USF3v3, sort_type=0)
        expected_keys = ('statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)

    def test_players_sort_by_type_with_sort_type_enum(self):
        response = self.api_client.leaderboard(leaderboard_id=cohpy.Codes.USF3v3,
                                               sort_type=SortType.ELO)
        expected_keys = ('statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)

    def test(self):
        response = self.api_client.leaderboard(leaderboard_id=cohpy.Codes.USF3v3,
                                               sort_type=SortType.ELO,
                                               start=200)
        expected_keys = ('statGroups', 'leaderboardStats', 'rankTotal',)

        assert all(key in expected_keys for key in response)
